#include "xio.h"
#include "xbasic_types.h"

#ifndef BYTE
#define BYTE unsigned char
#endif
#ifndef WORD
#define WORD unsigned short
#endif

// NE2000 definitions
#define NIC_BASE (0x00A00400) // Base I/O address of the NIC card
#define DATAPORT (0x10*2)
#define NE_RESET (0x1f*2)


// NIC page0 register offsets
#define CMDR    (0x00*2) // command register for read & write
#define PSTART  (0x01*2) // page start register for write
#define PSTOP   (0x02*2) // page stop register for write
#define BNRY    (0x03*2) // boundary reg for rd and wr
#define TPSR    (0x04*2) // tx start page start reg for wr
#define TBCR0   (0x05*2) // tx byte count 0 reg for wr
#define TBCR1   (0x06*2) // tx byte count 1 reg for wr
#define ISR     (0x07*2) // interrupt status reg for rd and wr
#define RSAR0   (0x08*2) // low byte of remote start addr
#define RSAR1   (0x09*2) // hi byte of remote start addr
#define RBCR0   (0x0A*2) // remote byte count reg 0 for wr
#define RBCR1   (0x0B*2) // remote byte count reg 1 for wr
#define RCR     (0x0C*2) // rx configuration reg for wr
#define TCR     (0x0D*2) // tx configuration reg for wr
#define DCR     (0x0E*2) // data configuration reg for wr
#define IMR     (0x0F*2) // interrupt mask reg for wr

// NIC page 1 register offsets
#define PAR0    (0x01*2) // physical addr reg 0 for rd and wr
#define CURR    (0x07*2) // current page reg for rd and wr
#define MAR0    (0x08*2) // multicast addr reg 0 for rd and WR

// Buffer Length and Field Definition Info
#define TXSTART  0x40              // Tx buffer start page
#define TXPAGES  6                 // Pages for Tx buffer
#define RXSTART  (TXSTART+TXPAGES) // Rx buffer start page
#define RXSTOP   0x7e              // Rx buffer end page for word mode
#define DCRVAL   0x49              // DCR values for word mode

// macros for reading and writting registers
#define outnic(addr, data) XIo_Out16(NIC_BASE+addr, data)
#define innic(addr) XIo_In16(NIC_BASE+addr)

// Private prototypes
void diag();
int init_etherne();
void resetnic();
void dmaRead(void *data, WORD addr, WORD length);
void dmaWrite(WORD *data, WORD addr, WORD length);
void send(BYTE *packet, WORD length);
//WORD innic(int reg);
//void outnic(int reg, int b);
void delay(int mult);

void startTX(WORD length);
inline void addTXData(WORD data);
void stopTX();

// packet prototype
static BYTE packet[1400];

int main(){
  int i;
  WORD *pkt;
  WORD word;

  microblaze_enable_icache();

  print("\r\n\r\n---------------Starting-------------\r\n\r\n");
  diag();

  // reset the NIC card
  if(!init_etherne()){
    print("Ethernet NIC not detected");
    return;
  }
  
  // create a broadcast packet
  packet[0] = 0xff;
  packet[1] = 0xff;
  packet[2] = 0xff;
  packet[3] = 0xff;
  packet[4] = 0xff;
  packet[5] = 0xff;

  // source addr
  packet[6] = 0x01;
  packet[7] = 0x02;
  packet[8] = 0x03;
  packet[9] = 0x04;
  packet[10] = 0x05;
  packet[11] = 0x06;

  // length field
  packet[12] = 0x6a;
  packet[13] = 0x05;

  for(i=14; i<1400; i++)
    packet[i] = (BYTE)i;

  print("\r\n");
  print("Sending Packets...\r\n");
  
  pkt = (WORD *)packet;
  while(1){
    startTX(1400);
    for(i=0; i<700; i++){
      word = (pkt[i]<<8)|(pkt[i]>>8);
      outnic(DATAPORT, word);
    }
    stopTX();
    //send(&packet, 1400);
  }

  print("\r\n\r\n---------------Done-------------\r\n");
  return 0;
}

// diagnostic that tests nic functionality
void diag(){
  int i;
  
  print("\r\n**************** Beginning of Diagnostic ****************\r\n\r\n");

  print("         Command Register Page Switching Test\r\n");
  print("Switching to page 1\r\n");
  outnic(CMDR, 0x61);

  print("Writing to and reading from 0x0D (value should be 0x4e): ");
  outnic(0x0D*2,0x4e);
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 0\r\n");
  outnic(CMDR, 0x21);

  print("Reading from reg offset 0x0D: ");
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 1\r\n");
  outnic(CMDR, 0x61);
  print("Reading from reg offset 0x0D (should be 0x4e): ");
  putnum(innic(0x0D*2));
  print("\r\n\r\n");

  print("         Default Value Test\r\n");
  print("Switching to page 0\r\n");
  outnic(CMDR,21);
  print("Reading from 0x16 (value should be 0x15): ");
  putnum(innic(0x16*2));
  print("\r\n");
  print("Reading from 0x12 (value should be 0x0c): ");
  putnum(innic(0x12*2));
  print("\r\n");
  print("Reading from 0x13 (value should be 0x12): ");
  putnum(innic(0x13*2));
  print("\r\n\r\n**************** End of Diagnostic ****************\r\n\r\n");
}

int init_etherne()
{
  outnic(NE_RESET, innic(NE_RESET));  // Do reset
  delay(2);
  if ((innic(ISR) & 0x80) == 0)       // Report if failed
  {
    print("  Ethernet card failed to reset!\r\n");
    return 0;
  }
  else
  {
    print("Ethernet card reset successful...");
    resetnic(); // Reset Ethernet card,
    print("Ethernet card intialization complete!\r\n");
  }
  return 1;
}

void resetnic(){
  int i;

  outnic(CMDR, 0x21);  // Abort and DMA and stop the NIC
  delay(2);

  outnic(DCR, DCRVAL);   // Set word-wide access

  outnic(RBCR0, 0x00); // Clear the count regs
  outnic(RBCR1, 0x00);

  outnic(IMR, 0x00);   // Mask completion irq
  outnic(ISR, 0xFF);   // clear interrupt status register

  outnic(RCR, 0x20);   // 0x20  Set to monitor mode

  // Set Rx start, Rx stop, Boundary and TX start regs
  outnic(PSTART, RXSTART);
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));
  outnic(TPSR, TXSTART);
  
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(IMR, 0x00);   // Mask completion irq

  outnic(CMDR, 0x61);

  // put nic in start mode
  outnic(CMDR, 0x22); // start nic
  outnic(TCR, 0x00);  // put nic in normal operation
}

void send(BYTE *packet, WORD length){
  WORD *pkt = (WORD *)packet;

  while(innic(CMDR)&0x04); // block while the NIC is busy

  outnic(TPSR, TXSTART); // set Transmit Start Page
  
  dmaWrite(packet, TXSTART<<8, length); // do DMA write
  
  outnic(CMDR, 0x24); // start the tranmission
  outnic(ISR, 0xff);
}

void dmaWrite(WORD *data, WORD addr, WORD len){
  int i;
  WORD counter;
  BYTE odd = 0;

  counter = len>>1;

  /*
  // if odd number of bytes add another write to memory
  if((len&0x1) == 1){
    counter--;
    odd = 1;
  }
  */

  outnic(RSAR0, (addr&0xff)); // set DMA starting address
  outnic(RSAR1, (addr>>8));

  outnic(ISR, 0xff); // clear ISR

  outnic(RBCR0, (len&0xff)); // set Remote DMA Byte Count
  outnic(RBCR1, (len>>8));

  outnic(TBCR0, (len&0xff)); // set Transmit Byte Count
  outnic(TBCR1, (len>>8));

  outnic(CMDR, 0x12); // start the DMA write

  // change order of MS/LS since DMA
  // writes LS byte in 15-8, and MS byte in 7-0
  for(i=0; i<counter; i++){
    outnic(DATAPORT, (data[i]<<8|data[i]>>8));//data[i+1]<<8)|data[i]); // write data
  }

  /*
  // send out last byte if the length is odd
  if(odd){
    outnic(DATAPORT, data[counter]); // write data
  }
  */
}

void dmaRead(void *data, WORD addr, WORD len){
  int i;
  int value;

  // since RBCR goes down to zero, set it equal to len-1
  // when using even number bytes
  // for odd number bytes leave it alone and the last 16-bit
  // read will include the last byte in the upper 8 bits.
  if((len&0x01) == 0){ // even number of bytes
    len--;
  }
  len+=10;
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(RBCR0, (len&0xff));
  outnic(RBCR1, (len>>8));
  outnic(RSAR0, 0x3300 | (addr&0xff)); // DMA starting address
  outnic(RSAR1, 0x4400 | (addr>>8));
  outnic(CMDR, 0x08);  // start the DMA read
  delay(2);
  for(i=0; i<=len/2; i++){
    innic(DATAPORT);
  }
}

/*// Input a word from a NIC register
WORD innic(int reg){
  return(XIo_In16(NIC_BASE+reg));
}

// Output a word to a NIC register
void outnic(int reg, int b){
  XIo_Out16((NIC_BASE+reg), b);
}
*/
void delay(int mult){
  int i;
  int delay = 1000000*mult;
  for(i=0; i<delay; i++);
}

void startTX(WORD len){
  while(innic(CMDR)&0x04); // block while the NIC is busy

  outnic(TPSR, TXSTART); // set Transmit Start Page
  
  outnic(RSAR0, (0x00)); // set DMA starting address
  outnic(RSAR1, (TXSTART));

  outnic(ISR, 0xff); // clear ISR

  outnic(RBCR0, (len&0xff)); // set Remote DMA Byte Count
  outnic(RBCR1, (len>>8));

  outnic(TBCR0, (len&0xff)); // set Transmit Byte Count
  outnic(TBCR1, (len>>8));  
  outnic(CMDR, 0x12); // start the DMA write
}

inline void addTXData(WORD data){
  // change order of MS/LS since DMA
  // writes LS byte in 15-8, and MS byte in 7-0
  outnic(DATAPORT, (data<<8|data>>8));
}

void stopTX(){
  outnic(CMDR, 0x24); // start the tranmission
  outnic(ISR, 0xff);
}
