#include "xio.h"
#include "xbasic_types.h"

#ifndef BYTE
#define BYTE unsigned char
#endif
#ifndef WORD
#define WORD unsigned short
#endif
#ifndef LWORD
#define LWORD unsigned long
#endif

// Ensure network structures aren't padded (DJGPP and VC)
#pragma pack(1)

#define MACLEN      6           // Ethernet (MAC) address length
#define BCASTADDR 0xff,0xff,0xff,0xff,0xff,0xff // Broadcast address

/* Ethernet Frame sizes */
#define MAXFRAME    1514 // Maximum frame size (inc hdr)
#define MINFRAME    64   // Minimum frame size (inc hdr)
#define ETHERMTU    80   // (MAXFRAME-sizeof(ETHERHDR))

// Ethernet (DIX) header
typedef struct {
  BYTE dest[MACLEN]; // Destination MAC address
  BYTE srce[MACLEN]; // Source MAC address
  WORD ptype;        // Protocol type or length
} ETHERHDR;

// Ethernet frame; data size is frame size minus header & CRC
typedef struct {
  ETHERHDR h;          //Header
  BYTE data[ETHERMTU]; // Data
} ETHERFRAME;

// NE2000-compatible driver prototypes
int init_etherne();
void poll_etherne();
WORD get_etherne(void *pkt);
WORD put_etherne(void *pkt, WORD len);

#define WORDMODE 1 // Set to zero if using 8-bit bus

// NE2000 definitions
#define NIC_BASE (0x00A00400) // Base I/O address of the NIC card
#define DATAPORT (0x10*2)
#define NE_RESET (0x1f*2)


// NIC page0 register offsets
#define CMDR    (0x00*2) // command register for read & write
#define PSTART  (0x01*2) // page start register for write
#define PSTOP   (0x02*2) // page stop register for write
#define BNRY    (0x03*2) // boundary reg for rd and wr
#define TPSR    (0x04*2) // tx start page start reg for wr
#define TBCR0   (0x05*2) // tx byte count 0 reg for wr
#define TBCR1   (0x06*2) // tx byte count 1 reg for wr
#define ISR     (0x07*2) // interrupt status reg for rd and wr
#define RSAR0   (0x08*2) // low byte of remote start addr
#define RSAR1   (0x09*2) // hi byte of remote start addr
#define RBCR0   (0x0A*2) // remote byte count reg 0 for wr
#define RBCR1   (0x0B*2) // remote byte count reg 1 for wr
#define RCR     (0x0C*2) // rx configuration reg for wr
#define TCR     (0x0D*2) // tx configuration reg for wr
#define DCR     (0x0E*2) // data configuration reg for wr
#define IMR     (0x0F*2) // interrupt mask reg for wr

// NIC page 1 register offsets
#define PAR0    (0x01*2) // physical addr reg 0 for rd and wr
#define CURR    (0x07*2) // current page reg for rd and wr
#define MAR0    (0x08*2) // multicast addr reg 0 for rd and WR

// Buffer Length and Field Definition Info
#define TXSTART  0x40              // Tx buffer start page
#define TXPAGES  6                 // Pages for Tx buffer
#define RXSTART  (TXSTART+TXPAGES) // Rx buffer start page
#if WORDMODE
#define RXSTOP   0x7e              // Rx buffer end page for word mode
#define DCRVAL   0x49              // DCR values for word mode
#else
#define RXSTOP   0x5f              // Ditto for byte mode
#define DCRVAL   0x48
#endif
#define STARHACK 0                  // Set non-zero to enable Starlan length hack


typedef struct        // Net driver configuration data
{
  BYTE myeth[MACLEN]; // MAC (Ethernet) addr
  WORD next_pkt;      // Next (current) Rx page
} CONFIGNE;

static CONFIGNE nic; // driver condifuration data
int promisc=0;       // Flag to enable promiscuous mode

typedef struct {     // NIC hardware packet header
  BYTE stat;         // Error status
  BYTE next;         // Pointer to next block
  WORD len;          // Length of this frame incl. CRC
} NICHDR;
NICHDR nichdr;

// Private prototypes
void resetnic(char cold);
void reset2();
void reset3();
void getnic(WORD addr, BYTE data[], WORD len);
void putnic(WORD addr, BYTE data[], WORD len);
BYTE nicwrap(int page);
BYTE innic(int reg);
void outnic(int reg, int b);
void delay(int mult);
void dmaRead(void *data, WORD addr, WORD length);
void dmaWrite(void *data, WORD addr, WORD length);

BYTE pkt[80];

int main(){
  int ok=0;

  print("\r\n\r\nStarting...\r\n");
  diag();
  if(!init_etherne()){
    print("Ethernet NIC not detected");
    return;
  }

  print("\r\nTransmitting Data\r\n");
  
  // destination address
  pkt[0] = 0xff;
  pkt[1] = 0xff;
  pkt[2] = 0xff;
  pkt[3] = 0xff;
  pkt[4] = 0xff;
  pkt[5] = 0xff;

  // source address
  pkt[6] = 64;
  pkt[7] = 12;
  pkt[8] = 98;
  pkt[9] = 98;
  pkt[10] = 12;
  pkt[11] = 64;

  pkt[12] = 0x08;
  pkt[13] = 0x00;

  for(ok=14; ok<80; ok++){
    pkt[ok] = (BYTE) ok;
  }
  
  put_etherne(pkt,80);

  print("...Done!\r\n");
  return 0;
}

// diagnostic that tests nic functionality
void diag(){
  int i;

  print("\r\n**************** Beginning of Diagnostic ****************\r\n\r\n");

  print("         Command Register Test\r\n");
  print("Switching to page 1\r\n");
  outnic(CMDR, 0x60);

  print("Writing to and reading from 0x0D (value should be 0x4e): ");
  outnic(0x0D*2,0x4e);
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 0\r\n");
  outnic(CMDR, 0x21);

  print("Reading from reg offset 0x0D: ");
  putnum(innic(0x0D*2));
  print("\r\n");

  print("Switching to page 1\r\n");
  outnic(CMDR, 0x60);
  print("Reading from reg offset 0x0D (should be 0x4e): ");
  putnum(innic(0x0D*2));
  print("\r\n\r\n");

  print("         Default Value Test\r\n");
  print("Switching to page 0\r\n");
  outnic(CMDR,21);
  print("Reading from 0x16 (value should be 0x15): ");
  putnum(innic(0x16*2));
  print("\r\n");
  print("Reading from 0x12 (value should be 0x0c): ");
  putnum(innic(0x12*2));
  print("\r\n");
  print("Reading from 0x13 (value should be 0x12): ");
  putnum(innic(0x13*2));
  print("\r\n\r\n");

  print("         Reset Test\r\n");
  outnic(NE_RESET, innic(NE_RESET));
  if ((innic(ISR) & 0x80) == 0){
    print("  Ethernet card failed to reset!");
  }
  else{
    print("Ethernet card reset successful!");
  }
  print("\r\n\r\n**************** End of Diagnostic ****************\r\n");
}

int init_etherne()
{
  int ok=0;

  outnic(NE_RESET, innic(NE_RESET));  // Do reset
  delay(2);
  if ((innic(ISR) & 0x80) == 0)       // Report if failed
  {
    print("  Ethernet card failed to reset!\r\n");
  }
  else
  {
    print("Ethernet card reset successful\r\n");
    reset3();                // Reset Ethernet card, get my addr
    ok = 1;
  }
  return(ok);
}

void reset3(){
  int i;

  outnic(CMDR, 0x21);  // Abort and DMA and stop the NIC
  delay(2);
  outnic(DCR, 0x49);   // Set word-wide access
  outnic(RBCR0, 0x00); // Clear the count regs
  outnic(RBCR1, 0x00);
  outnic(IMR, 0x00);   // Mask completion irq
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(RCR, 0x20);   // 0x20  Set to monitor
  outnic(TCR, 0x02);   // 0x02  and loopback mode

  dmaWrite(&i, 0x0000, 32);
  dmaRead(&i, 0x0000, 32);

  outnic(TPSR, TXSTART);              // Set Tx start page
  outnic(PSTART, RXSTART);            // Set Rx start, stop, boundary
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));

  outnic(ISR, 0xff);
  outnic(TCR, 0x00);
  outnic(CMDR, 0x22);

  dmaWrite(&i, 0x0040, 32);
  dmaRead(&i, 0x0040, 32);
  /*outnic(RBCR0, 0x1f);
  outnic(RBCR1, 0x00);
  outnic(RSAR0, 0x00); // DMA starting at 0x0000
  outnic(RSAR1, 0x40);
  outnic(CMDR, 0x0a);  // start the DMA read
  print("\r\n");
  for(i=0; i<16; i++){
    putnum(innic(DATAPORT));
    print(" ");
    putnum(innic(8*2));
    print(" ");
    putnum(innic(9*2));
    print(" ");
    putnum(innic(ISR));
    print("\r\n");
    }*/
}

void dmaRead(void *data, WORD addr, WORD len){
  int i;
  len--;
  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(RBCR0, len&0xff);
  outnic(RBCR1, len>>8);
  outnic(RSAR0, addr&0xff); // DMA starting address
  outnic(RSAR1, addr>>8);
  outnic(CMDR, 0x0a);  // start the DMA read
  delay(2);
  for(i=0; i<16; i++){
    putnum(innic(DATAPORT));
    print(" ");
    putnum(innic(8*2));
    print(" ");
    putnum(innic(9*2));
    print(" ");
    putnum(innic(ISR));
    print("\r\n");
  }
}

void dmaWrite(void *data, WORD addr, WORD len){
  int i;
  len--;
  outnic(RBCR0, 0x42);
  outnic(RBCR1, 0x00);
  outnic(RSAR0, 0x42);
  outnic(RSAR1, 0x00);
  outnic(CMDR, 0x0a);

  outnic(ISR, 0xFF);   // clear interrupt status register
  outnic(RBCR0, len&0xff);
  outnic(RBCR1, len>>8);
  outnic(RSAR0, addr&0xff); // DMA starting address
  outnic(RSAR1, addr>>8);
  outnic(CMDR, 0x12);  // start the DMA write
  delay(2);
  for(i=0; i<16; i++){
    outnic(DATAPORT,0xee);
    print(" ");
    putnum(innic(8*2));
    print(" ");
    putnum(innic(9*2));
    print(" ");
    putnum(innic(ISR));
    print("\r\n");
  }
}

// Initialise card given driver type and base addr.
void reset2(){
  int i, j;
  register int count;

  outnic(CMDR, 0x21);   // Stop, DMA abort, page 0
  delay(5);             // ..wait to take effect
  outnic(DCR, DCRVAL);
  outnic(RBCR0, 0);     // Clear remote byte count
  outnic(RBCR1, 0);
  outnic(IMR, 0x00);    // Mask interrupts  
  outnic(ISR, 0xff);    // Clear interrupt flags
  outnic(RCR, 0x20);    // Rx monitor mode
  outnic(TCR, 0x02);    // Tx internal loopback

  outnic(TPSR, TXSTART);              // Set Tx start page
  outnic(PSTART, RXSTART);            // Set Rx start, stop, boundary
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));

  outnic(RBCR0, 32);    // Set remote byte count
  outnic(RBCR1, 0);
  outnic(RSAR0, 0);                   
  outnic(RSAR1, 0);

  outnic(CMDR,0x0a);

  print("         ");
  putnum(innic(0x08*2));
  print(" ");
  putnum(innic(0x09*2));
  print("\r\n");
  
  for(i=0; i<16; i++){
    putnum(innic(DATAPORT));
    print(" ");
    putnum(innic(0x08*2));
    print(" ");
    putnum(innic(0x09*2));
    print("\r\n");
    if(innic(ISR) != 0)
      break;
  }

  count = 100000;
  while(count && (innic(ISR)&0x40)==0)
    count--;
  print("count: ");
  putnum(count);
  print(" ");
  putnum(innic(ISR));
  print("\r\n");

  outnic(RCR, promisc ? 0x14 : 0x04); // Allow broadcasts, maybe all pkts
  outnic(TCR, 0);                     // Normal Tx operation
  outnic(ISR, 0xff);                  // Clear interrupt flags
  outnic(CMDR, 0x22);                 // Start NIC
}

// Reset the Ethernet card, if 'cold' start, get my 6-byte address
void resetnic(char cold){
  int i;
  BYTE temp[MACLEN*2];
  
  outnic(CMDR, 0x21);                 // Stop, DMA abort, page 0
  delay(5);                           // ..wait to take effect
  outnic(DCR, DCRVAL);
  outnic(RBCR0, 0);                   // Clear remote byte count
  outnic(RBCR1, 0);
  outnic(RCR, 0x20);                  // Rx monitor mode
  outnic(TCR, 0x02);                  // Tx internal loopback
  outnic(TPSR, TXSTART);              // Set Tx start page
  outnic(PSTART, RXSTART);            // Set Rx start, stop, boundary
  outnic(PSTOP, RXSTOP);
  outnic(BNRY, (BYTE)(RXSTOP-1));
  outnic(ISR, 0xff);                  // Clear interrupt flags
  outnic(IMR, 0x00);                     // Mask all interrupts

  outnic(RBCR0, 32);                   // Clear remote byte count
  outnic(RBCR1, 0);
  outnic(RSAR0, 0);                   
  outnic(RSAR1, 0);

  if (cold)
  {
    outnic(CMDR, 0x22);             // Start NIC, DMA abort
    getnic(0, temp, 12);            // Get 6-byte addr
    for (i=0; i<MACLEN; i++){        // Convert addr words to bytes
      nic.myeth[i] = temp[WORDMODE ? i+i : i];
    }
  }
  outnic(CMDR, 0x61);                 // Stop, DMA abort, page 1
  delay(2);
  for (i=0; i<6; i++)                 // Set Phys addr
    outnic(PAR0+i, nic.myeth[i]);
  for (i=0; i<8; i++){                 // Multicast accept-all
    outnic(MAR0+(i*2), 0xef);
  }

  outnic(CURR, RXSTART+1);            // Set current Rx page
  nic.next_pkt = RXSTART + 1;
  outnic(CMDR, 0x20);                 // DMA abort, page 0
  outnic(RCR, promisc ? 0x14 : 0x04); // Allow broadcasts, maybe all pkts
  outnic(TCR, 0);                     // Normal Tx operation
  outnic(ISR, 0xff);                  // Clear interrupt flags
  outnic(CMDR, 0x22);                 // Start NIC
}

// Send Ethernet packet given len, return 0 if NIC is busy
WORD put_etherne(void *pkt, WORD len)
{
  int i;
  register int count;

  if (innic(CMDR) & 0x4)      // If still Txing, return 0
    len = 0;
  else if (pkt)
  {                                 // If last Tx is complete..
    outnic(ISR, 0xff);              // Clear interrupt flags
    outnic(TBCR0, len & 0xff);      // Set Tx length regs
    outnic(TBCR1, len >> 8);
    putnic(TXSTART<<8, pkt, len<<1);
    outnic(ISR, 0xff);              // Clear interrupt flags
    outnic(CMDR, 0x26);             // Transmit the packet
    count = 100000;                      // Done: must ensure DMA complete
    while(count && (i=innic(ISR))==0)
      count--;
    print("count: ");
    putnum(count);
    print(" " );
    putnum(innic(ISR));
    print("\r\n");
  }
  return(len);
}

// Put a packet into a given address in the NIC's RAM
void putnic(WORD addr, BYTE data[], WORD len){
  int i;
  register int count;
  register WORD *dataw, dataport;
  
  len += len & 1;                     // Round length up to an even value
  count = WORDMODE ? len>>1 : len;    // Halve byte count if word O/P
  outnic(ISR, 0xff);                  // Clear remote DMA interrupt flag
  outnic(RBCR0, len&0x40);            // Byte count
  outnic(RBCR1, len>>8);
  outnic(RSAR0, addr&0xff);           // Data addr
  outnic(RSAR1, addr>>8);
  outnic(CMDR, 0x12);                 // Start, DMA remote write
#if WORDMODE                          // Word transfer?
  dataw = (WORD *)data;
  while(count--)
    outnic(DATAPORT, *dataw++);       // O/P words
#else
  while(count--)                      // O/P bytes
    outnic(DATAPORT, *data++);
#endif
  count = 100000;                      // Done: must ensure DMA complete
  while(count && (i=innic(ISR))==0)
    count--;
  print("count: ");
  putnum(count);
  print(" " );
  putnum(i);
  print("\r\n");
}

// Get a packet from a given address in the NIC's RAM
void getnic(WORD addr, BYTE data[], WORD len){
    register int count;
    register WORD *dataw, dataport;

    count = WORDMODE ? len>>1 : len;    // Halve byte count if word I/P
    outnic(ISR, 0x40);                  // Clear remote DMA interrupt flag
    outnic(RBCR0, len&0xff);            // Byte count
    outnic(RBCR1, len>>8);
    outnic(RSAR0, addr&0xff);           // Data addr
    outnic(RSAR1, addr>>8);
    outnic(CMDR, 0x0a);                 // Start, DMA remote read

#if WORDMODE
    dataw = (WORD *)data;               // Use pointer for speed
    while(count--)                      // Get words
        *dataw++ = innic(DATAPORT);
    if (len & 1)                        // If odd length, do last byte
        *(BYTE *)dataw = innic(DATAPORT);
#else
    while(count--)                      // Get bytes
        *data++ = innic(DATAPORT);
#endif
    putnum(innic(ISR));

}

// Poll network interface to keep it alive; send & receive frames
void poll_etherne(){
  WORD len;
  static BYTE ebuff[MAXFRAME];

  outnic(ISR, 0x01);                  // Clear interrupt flag
  /*
  // Receive
  while ((len=get_etherne(ebuff))>0)
  {                                   // Store frames in buff
    receive_upcall(ebuff, len);
  }
  // Transmit
  while (!(innic(CMDR)&0x04) &&       // While NIC ready & frame avail
	 (len=transmit_upcall(ebuff, MAXFRAME))>0)
  {                                   // ..transmit frame
    put_etherne(ebuff, len);
  }
  */
}

/*
// Get packet into buffer, return length (excl CRC), or 0 if none available
WORD get_etherne(void *pkt){
  WORD len=0, curr;
  BYTE bnry;
  CONFIGNE *cp;
  
  cp = &nic;
  if (innic(ISR) & 0x10)              // If Rx overrun..
  {
    print("  NIC Rx overrun\r\n");
    resetnic(cp, 0);                // ..reset controller (drastic!)
  }
  outnic(CMDR, 0x60);                 // DMA abort, page 1
  curr = innic(CURR);                 // Get current page
  outnic(CMDR, 0x20);                 // DMA abort, page 0
  if (curr != cp->next_pkt)           // If Rx packet..
  {
    nichdr.len = 0xee;
    getnic((WORD)(cp->next_pkt<<8), (BYTE *)&nichdr, sizeof(nichdr));
    
    len = nichdr.len;               // Take length from stored header
    if ((nichdr.stat&1) && len>=MINFRAME && len<=MAXFRAME)
    {                               // If hdr is OK, get packet
      if (pkt)
	getnic((WORD)((cp->next_pkt<<8)+sizeof(nichdr)), pkt, len);
    }
    else                            // If not, no packet data
    {
      print("  NIC packet error\r\n");
    }                               // Update next packet ptr
    if (nichdr.next>=RXSTART && nichdr.next<RXSTOP)
      cp->next_pkt = nichdr.next;
    else                            // If invalid, use prev+1
    {
      print("  NIC pointer error\r\n");
      cp->next_pkt = nicwrap(cp->next_pkt + 1);
    }                               // Update boundary register
    bnry = nicwrap(cp->next_pkt - 1);
    outnic(BNRY, bnry);
  }
  return(len);                        // Return length excl. CRC
}
*/

// Wrap an NIC Rx page number
BYTE nicwrap(int page){
  if (page >= RXSTOP)
    page += RXSTART - RXSTOP;
  else if (page < RXSTART)
    page += RXSTOP - RXSTART;
  return(page);
}

// Input a byte from a NIC register
BYTE innic(int reg){
  return(XIo_In16(NIC_BASE+reg));
}

// Output a byte to a NIC register
void outnic(int reg, int b){
  XIo_Out16((NIC_BASE+reg), b);
}

void delay(int mult){
  int i;
  int delay = 1000000*mult;
  for(i=0; i<delay; i++);
}
